from datetime import timedelta

import responses
import swapper
from django.core.exceptions import ValidationError
from django.test import TestCase
from freezegun import freeze_time
from netdiff import OlsrParser

from .utils import CreateGraphObjectsMixin, CreateOrgMixin, LoadMixin

Link = swapper.load_model('topology', 'Link')
Node = swapper.load_model('topology', 'Node')
Topology = swapper.load_model('topology', 'Topology')


class TestTopology(CreateOrgMixin, CreateGraphObjectsMixin, LoadMixin, TestCase):
    topology_model = Topology
    node_model = Node
    link_model = Link
    maxDiff = None

    def setUp(self):
        org = self._create_org()
        t = self._create_topology(organization=org)
        self._create_node(
            label='node1', addresses=['192.168.0.1'], topology=t, organization=org
        )
        self._create_node(
            label='node2', addresses=['192.168.0.2'], topology=t, organization=org
        )

    def _get_nodes(self):
        return self.node_model.objects.all()

    def _set_receive(self, expiration_time=0, parser='netdiff.NetJsonParser'):
        t = self.topology_model.objects.first()
        t.parser = parser
        t.strategy = 'receive'
        t.key = 'test'
        t.expiration_time = expiration_time
        t.save()
        return t

    def test_str(self):
        t = self.topology_model.objects.first()
        self.assertIsInstance(str(t), str)

    def test_parser(self):
        t = self.topology_model.objects.first()
        self.assertIs(t.parser_class, OlsrParser)

    def test_json_empty(self):
        t = self.topology_model.objects.first()
        self.node_model.objects.all().delete()
        graph = t.json(dict=True)
        self.assertDictEqual(
            graph,
            {
                'type': 'NetworkGraph',
                'protocol': 'OLSR',
                'version': '0.8',
                'metric': 'ETX',
                'label': t.label,
                'id': str(t.id),
                'parser': t.parser,
                'created': t.created,
                'modified': t.modified,
                'nodes': [],
                'links': [],
            },
        )

    def test_json(self):
        node1, node2 = self._get_nodes()
        t = self.topology_model.objects.first()
        node3 = t._create_node(addresses=['192.168.0.3'], label='node3')
        node3.save()
        link = t._create_link(source=node1, target=node2, cost=1)
        link.save()
        l2 = t._create_link(source=node1, target=node3, cost=1)
        l2.save()
        graph = t.json(dict=True)
        self.assertDictEqual(
            dict(graph),
            {
                'type': 'NetworkGraph',
                'protocol': 'OLSR',
                'version': '0.8',
                'metric': 'ETX',
                'label': t.label,
                'id': str(t.id),
                'parser': t.parser,
                'created': t.created,
                'modified': t.modified,
                'nodes': [
                    dict(node1.json(dict=True)),
                    dict(node2.json(dict=True)),
                    dict(node3.json(dict=True)),
                ],
                'links': [dict(link.json(dict=True)), dict(l2.json(dict=True))],
            },
        )
        self.assertIsInstance(t.json(), str)

    @responses.activate
    def test_empty_diff(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=t.json(original=True),
            content_type='application/json',
        )
        self.assertDictEqual(
            t.diff(), {'added': None, 'removed': None, 'changed': None}
        )

    @responses.activate
    def test_update_all_attributes(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=self._load('static/netjson-1-link.json'),
            content_type='application/json',
        )
        t.protocol = None
        t.version = None
        t.metric = None
        t.update()
        t.refresh_from_db()
        self.assertEqual(t.protocol, 'OLSR')
        self.assertEqual(t.version, '0.8')
        self.assertEqual(t.metric, 'ETX')

    @responses.activate
    def test_update_added(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=self._load('static/netjson-1-link.json'),
            content_type='application/json',
        )
        self.node_model.objects.all().delete()
        t.update()
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)
        node1 = self.node_model.objects.get(addresses__contains='"192.168.0.1"')
        node2 = self.node_model.objects.get(addresses__contains='"192.168.0.2"')
        self.assertEqual(node1.local_addresses, ['10.0.0.1'])
        self.assertEqual(node1.properties, {'gateway': True})
        link = self.link_model.objects.first()
        self.assertIn(link.source, [node1, node2])
        self.assertIn(link.target, [node1, node2])
        self.assertEqual(link.cost, 1.0)
        self.assertEqual(link.properties, {'pretty': True})
        # ensure repeating the action is idempotent
        t.update()
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)

    @responses.activate
    def test_update_changed(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=self._load('static/netjson-1-link.json'),
            content_type='application/json',
        )
        self.node_model.objects.all().delete()
        t.update()
        node = self.node_model.get_from_address('192.168.0.2', topology=t)
        self.assertEqual(node.label, '')
        self.assertEqual(node.addresses, ['192.168.0.2'])
        link = self.link_model.objects.first()
        self.assertEqual(link.properties, {'pretty': True})
        self.assertEqual(link.cost, 1.0)
        self.assertEqual(link.cost_text, '10 Mbps')
        # now change
        t.url = t.url.replace('9090', '9091')
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9091',
            body=self._load('static/netjson-2-links.json'),
            content_type='application/json',
        )
        t.update()
        self.assertEqual(self.node_model.objects.count(), 3)
        # test changed properties for existing node
        node = self.node_model.get_from_address('192.168.0.2', topology=t)
        self.assertEqual(node.label, 'node2')
        self.assertEqual(node.addresses, ['192.168.0.2', '10.0.0.2'])
        # test added node with properties
        node = self.node_model.get_from_address('192.168.0.3', topology=t)
        self.assertEqual(node.label, 'node3')
        self.assertEqual(node.addresses, ['192.168.0.3', '10.0.0.3'])
        self.assertEqual(self.link_model.objects.count(), 2)
        link.refresh_from_db()
        self.assertEqual(link.properties, {'pretty': True})
        self.assertEqual(link.cost, 1.5)
        self.assertEqual(link.cost_text, '15 Mbps')
        link = self.link_model.get_from_nodes(
            source='192.168.0.1', target='192.168.0.3', topology=t
        )
        self.assertEqual(link.properties, {'pretty': False})
        self.assertEqual(link.cost, 2.0)
        self.assertEqual(link.cost_text, '20 Mbps')

    @responses.activate
    def test_update_removed(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=self._load('static/netjson-2-links.json'),
            content_type='application/json',
        )
        self.node_model.objects.all().delete()
        t.update()
        self.assertEqual(self.node_model.objects.count(), 3)
        self.assertEqual(self.link_model.objects.count(), 2)
        # now change
        t.url = t.url.replace('9090', '9091')
        t.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9091',
            body=self._load('static/netjson-1-link.json'),
            content_type='application/json',
        )
        t.update()
        self.assertEqual(self.node_model.objects.count(), 3)
        self.assertEqual(self.link_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.filter(status='down').count(), 1)
        link = self.link_model.objects.filter(status='down').first()
        self.assertIn('192.168.0.3', [link.source.netjson_id, link.target.netjson_id])
        self.assertEqual(link.cost, 2.0)

    @responses.activate
    def test_update_status_existing_link(self):
        t = self.topology_model.objects.first()
        t.parser = 'netdiff.NetJsonParser'
        t.save()
        n1 = self.node_model.objects.all()[0]
        n2 = self.node_model.objects.all()[1]
        link = t._create_link(
            source=n1, target=n2, cost=1, status='down', properties={'pretty': True}
        )
        link.full_clean()
        link.save()
        responses.add(
            responses.GET,
            'http://127.0.0.1:9090',
            body=self._load('static/netjson-1-link.json'),
            content_type='application/json',
        )
        t.update()
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)
        link.refresh_from_db()
        self.assertEqual(link.status, 'up')

    def test_topology_url_empty(self):
        t = self.topology_model(
            label='test', parser='netdiff.NetJsonParser', strategy='fetch'
        )
        with self.assertRaises(ValidationError):
            t.full_clean()

    def test_topology_key_empty(self):
        t = self.topology_model(
            label='test', parser='netdiff.NetJsonParser', strategy='receive', key=''
        )
        with self.assertRaises(ValidationError):
            t.full_clean()

    def _test_receive_added(self, expiration_time=0):
        self.node_model.objects.all().delete()
        t = self._set_receive(expiration_time)
        data = self._load('static/netjson-1-link.json')
        t.receive(data)
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)
        node1 = self.node_model.objects.get(addresses__contains='"192.168.0.1"')
        node2 = self.node_model.objects.get(addresses__contains='"192.168.0.2"')
        self.assertEqual(node1.local_addresses, ['10.0.0.1'])
        self.assertEqual(node1.properties, {'gateway': True})
        link = self.link_model.objects.first()
        self.assertIn(link.source, [node1, node2])
        self.assertIn(link.target, [node1, node2])
        self.assertEqual(link.cost, 1.0)
        self.assertEqual(link.properties, {'pretty': True})
        # ensure repeating the action is idempotent
        t.receive(data)
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)

    def test_receive_added(self):
        self._test_receive_added()

    def _test_receive_changed(self, expiration_time=0):
        t = self._set_receive(expiration_time)
        self.node_model.objects.all().delete()
        data = self._load('static/netjson-1-link.json')
        t.receive(data)
        link = self.link_model.objects.first()
        # now change
        data = self._load('static/netjson-2-links.json')
        t.receive(data)
        link.refresh_from_db()
        self.assertEqual(self.node_model.objects.count(), 3)
        self.assertEqual(self.link_model.objects.count(), 2)
        self.assertEqual(link.cost, 1.5)

    def test_receive_changed(self):
        self._test_receive_changed()

    def test_receive_removed(self):
        t = self._set_receive()
        self.node_model.objects.all().delete()
        data = self._load('static/netjson-2-links.json')
        t.receive(data)
        self.assertEqual(self.node_model.objects.count(), 3)
        self.assertEqual(self.link_model.objects.count(), 2)
        # now change
        data = self._load('static/netjson-1-link.json')
        t.receive(data)
        self.assertEqual(self.node_model.objects.count(), 3)
        self.assertEqual(self.link_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.filter(status='down').count(), 1)
        link = self.link_model.objects.filter(status='down').first()
        self.assertIn('192.168.0.3', [link.source.netjson_id, link.target.netjson_id])
        self.assertEqual(link.cost, 2.0)

    def test_receive_removed_bug(self):
        t = self._set_receive(parser='netdiff.OpenvpnParser')
        self.node_model.objects.all().delete()
        data = self._load('static/openvpn.txt')
        t.receive(data)
        self.assertEqual(self.node_model.objects.count(), 4)
        self.assertEqual(self.link_model.objects.count(), 3)
        self.assertEqual(self.link_model.objects.filter(status='up').count(), 3)
        self.assertEqual(self.link_model.objects.filter(status='down').count(), 0)
        data = self._load('static/openvpn-0-links.txt')
        t.receive(data)
        self.assertEqual(self.node_model.objects.count(), 4)
        self.assertEqual(self.link_model.objects.count(), 3)
        self.assertEqual(self.link_model.objects.filter(status='down').count(), 3)
        self.assertEqual(self.link_model.objects.filter(status='up').count(), 0)
        t.refresh_from_db()
        netjson = t.json(dict=True)
        for link in self.link_model.objects.all():
            for attr in ['status', 'status_changed', 'created', 'modified']:
                with self.subTest(f'{attr} should not be saved in properties ({link})'):
                    self.assertNotIn(attr, link.properties)
        for link in netjson['links']:
            with self.subTest(f'status should be down ({link})'):
                self.assertEqual(link['properties']['status'], 'down')
            for attr in ['status_changed', 'created', 'modified']:
                with self.subTest(f'{attr} should be present in properties ({link})'):
                    self.assertIn(attr, link['properties'])
        for node in netjson['nodes']:
            for attr in ['created', 'modified']:
                with self.subTest(f'{attr} should be present in properties ({node})'):
                    self.assertIn(attr, node['properties'])

    def test_receive_status_existing_link(self):
        t = self._set_receive()
        n1 = self.node_model.objects.all()[0]
        n2 = self.node_model.objects.all()[1]
        link = t._create_link(
            source=n1, target=n2, cost=1, status='down', properties={'pretty': True}
        )
        link.full_clean()
        link.save()
        data = self._load('static/netjson-1-link.json')
        t.receive(data)
        self.assertEqual(self.node_model.objects.count(), 2)
        self.assertEqual(self.link_model.objects.count(), 1)
        link.refresh_from_db()
        self.assertEqual(link.status, 'up')

    def test_multiple_receive_added(self):
        self._test_receive_added(expiration_time=50)

    def test_multiple_receive_changed(self):
        self._test_receive_changed(expiration_time=50)

    def test_multiple_receive_removed(self):
        with freeze_time() as frozen_time:
            t = self._set_receive(expiration_time=0.1)
            data = self._load('static/netjson-2-links.json')
            for sleep_time in [0, 0.2, 0.1]:
                frozen_time.tick(timedelta(seconds=sleep_time))
                t.receive(data)
                self.assertEqual(self.node_model.objects.count(), 3)
                self.assertEqual(self.link_model.objects.count(), 2)
                self.assertEqual(
                    self.link_model.objects.filter(status='down').count(), 0
                )
            # receive change
            data = self._load('static/netjson-1-link.json')
            t.receive(data)
            self.assertEqual(self.node_model.objects.count(), 3)
            self.assertEqual(self.link_model.objects.count(), 2)
            # expiration_time has not expired
            self.assertEqual(self.link_model.objects.filter(status='down').count(), 0)
            # expiration_time has now expired for 1 link
            frozen_time.tick(timedelta(seconds=0.2))
            t.receive(data)
            self.assertEqual(self.link_model.objects.filter(status='down').count(), 1)
            link = self.link_model.objects.filter(status='down').first()
            self.assertIn(
                '192.168.0.3', [link.source.netjson_id, link.target.netjson_id]
            )
            self.assertEqual(link.cost, 2.0)

    def test_multiple_receive_split_network(self):
        def _assert_split_topology(self, topology):
            self.assertEqual(self.node_model.objects.count(), 4)
            self.assertEqual(self.link_model.objects.filter(status='up').count(), 2)
            self.assertEqual(self.link_model.objects.filter(status='down').count(), 0)
            link = self.link_model.get_from_nodes(
                '192.168.0.1', '192.168.0.2', topology
            )
            self.assertIsNotNone(link)
            self.assertEqual(link.status, 'up')
            self.assertEqual(link.cost, 1.0)
            link = self.link_model.get_from_nodes(
                '192.168.0.10', '192.168.0.20', topology
            )
            self.assertIsNotNone(link)
            self.assertEqual(link.status, 'up')
            self.assertEqual(link.cost, 1.1)

        with freeze_time() as frozen_time:
            self.node_model.objects.all().delete()
            t = self._set_receive(expiration_time=0.5)
            network1 = self._load('static/netjson-1-link.json')
            network2 = self._load('static/split-network.json')
            t.receive(network1)
            t.receive(network2)
            _assert_split_topology(self, t)
            for sleep_time in [0.1, 0.25, 0.3]:
                frozen_time.tick(timedelta(seconds=sleep_time))
                t.receive(network1)
                _assert_split_topology(self, t)
                frozen_time.tick(timedelta(seconds=sleep_time))
                t.receive(network2)
                _assert_split_topology(self, t)

    def test_very_long_addresses(self):
        """
        see https://github.com/netjson/django-netjsongraph/issues/6
        """
        t = self._set_receive()
        data = self._load('static/very-long-addresses.json')
        t.receive(data)
        n = self.node_model.get_from_address('2001:4e12:452a:1:172::10', t)
        self.assertEqual(len("".join(n.addresses)), 568)

    def test_save_snapshot(self):
        t = self._set_receive()
        t.save_snapshot()
        s = t.snapshot_set.model.objects.first()
        self.assertEqual(s.data, t.json())
        self.assertEqual(s.topology, t)
        t.save_snapshot()
        s = t.snapshot_set.model.objects.first()
        self.assertFalse(s.created == s.modified)

    def test_label_addition(self):
        t = self._set_receive(parser='netdiff.OpenvpnParser')
        t.save()
        t.node_set.all().delete()
        data = self._load('static/openvpn.txt')
        t.receive(data)
        self.assertEqual(t.node_set.count(), 4)
        labels = []
        for node in t.node_set.all():
            labels.append(node.label)
        self.assertIn('Syskrack', labels)
        self.assertIn('Kali-Matera', labels)
        self.assertIn('pomezia', labels)

    def test_issue_58(self):
        self.node_model.objects.all().delete()
        self.link_model.objects.all().delete()
        t = self._set_receive()
        data = self._load('static/issue-58.json')
        t.receive(data)
        self.assertEqual(t.node_set.all().count(), 13)
        self.assertEqual(t.link_set.all().count(), 11)

    def test_update_added_items_regression_74_first(self):
        t = self._set_receive(parser='netdiff.OpenvpnParser')
        t.save()
        node = t.node_set.first()
        items = {'nodes': [{'id': node.netjson_id}]}
        t._update_added_items(items)

    def test_update_added_items_regression_74_second(self):
        t = self._set_receive(parser='netdiff.OpenvpnParser')
        t.save()
        items = {'nodes': [{'id': 'bogus'}]}
        t._update_changed_items(items)

    def test_receive_regression(self):
        t = self._set_receive(parser='netdiff.OpenvpnParser', expiration_time=120)
        t.save()
        data = self._load('static/openvpn.txt')
        pre_existing_nodes = Node.objects.count()
        node1 = Node.objects.first()
        node2 = Node.objects.last()
        with freeze_time() as frozen_time:
            t.receive(data)
            self.assertEqual(t.node_set.count(), 4 + pre_existing_nodes)
            self.assertEqual(t.link_set.count(), 3)
            # create new link not included in the topology data
            link = self._create_link(
                source=node1, target=node2, status='up', topology=t, cost=1
            )
            self.assertEqual(link.status, 'up')
            modified = link.modified
            # after 61 seconds we receive new (empty) data
            # we expect the link status to remain unchanged
            # (since expiration_time has not passed yet)
            # but we do not expect the link to be modified
            frozen_time.tick(timedelta(seconds=61))
            t.receive(data)
            link.refresh_from_db()
            self.assertEqual(link.status, 'up')
            self.assertEqual(link.modified, modified)
            # hen the topology data is received again later
            # the expiration time is passed and we expect
            # the link to be flagged as down and modified
            frozen_time.tick(timedelta(seconds=61))
            t.receive(data)
            link.refresh_from_db()
            self.assertEqual(link.status, 'down')
            self.assertNotEqual(link.modified, modified)



from unittest import mock

import django
import swapper
from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.db import transaction
from django.test import TransactionTestCase
from django.urls import reverse
from django.utils.module_loading import import_string
from openwisp_controller.config.tests.utils import (
    CreateConfigTemplateMixin,
    TestVpnX509Mixin,
)

from openwisp_network_topology.tests.utils import (
    CreateGraphObjectsMixin,
    CreateOrgMixin,
)

from .base.models import logger as models_logger
from .base.models import trigger_device_checks_path

Node = swapper.load_model('topology', 'Node')
Link = swapper.load_model('topology', 'Link')
Topology = swapper.load_model('topology', 'Topology')
DeviceNode = swapper.load_model('topology_device', 'DeviceNode')
Device = swapper.load_model('config', 'Device')
Template = swapper.load_model('config', 'Template')
Vpn = swapper.load_model('config', 'Template')
Organization = swapper.load_model('openwisp_users', 'Organization')
Cert = swapper.load_model('pki', 'Cert')


class Base(
    TestVpnX509Mixin, CreateConfigTemplateMixin, CreateGraphObjectsMixin, CreateOrgMixin
):
    topology_model = Topology
    node_model = Node

    def _get_org(self):
        org = Organization.objects.first()
        if not org:
            return Organization.objects.create(name='default', slug='default')
        return org

    def _init_test_node(
        sel, topology, addresses=None, label='test', common_name=None, create=True
    ):
        if not addresses:
            addresses = ['netjson_id']
        node = Node(
            organization=topology.organization,
            topology=topology,
            label=label,
            addresses=addresses,
            properties={'common_name': common_name},
        )
        if create:
            node.full_clean()
            node.save()
        return node

    def _create_test_env(self, parser):
        organization = self._get_org()
        vpn = self._create_vpn(name='test VPN', organization=organization)
        self._create_template(
            name='VPN',
            type='vpn',
            vpn=vpn,
            config=vpn.auto_client(),
            default=True,
            organization=organization,
        )
        vpn2 = self._create_vpn(name='test VPN2', ca=vpn.ca, organization=organization)
        self._create_template(
            name='VPN2',
            type='vpn',
            vpn=vpn2,
            config=vpn.auto_client(),
            default=True,
            organization=organization,
        )
        device = self._create_device(organization=organization)
        config = self._create_config(device=device)
        topology = self._create_topology(organization=organization, parser=parser)
        cert = config.vpnclient_set.first().cert
        return topology, device, cert


class TestControllerIntegration(Base, TransactionTestCase):
    def test_auto_create_openvpn(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        expected_queries = 12 if django.VERSION[0:2] >= (3, 0) else 14
        self.assertEqual(DeviceNode.objects.count(), 0)
        with self.subTest('assert number of queries'):
            with self.assertNumQueries(expected_queries):
                node = self._init_test_node(topology, common_name=cert.common_name)
        self.assertEqual(DeviceNode.objects.count(), 1)
        device_node = DeviceNode.objects.first()
        self.assertEqual(device_node.device, device)
        self.assertEqual(device_node.node, node)

        with self.subTest('not run on save'):
            with mock.patch.object(transaction, 'on_commit') as on_commit:
                node.save()
                on_commit.assert_not_called()

    def test_auto_create_openvpn_failures(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')

        with self.subTest('common_name not present'):
            self._init_test_node(topology)
            self.assertFalse(DeviceNode.objects.exists())

        with self.subTest('cert does not exist'):
            self._init_test_node(topology, common_name='missing')
            self.assertFalse(DeviceNode.objects.exists())

        with self.subTest('exception during save'):
            with mock.patch.object(
                DeviceNode, 'save', side_effect=Exception('test')
            ) as save:
                with mock.patch.object(models_logger, 'exception') as logger_exception:
                    self._init_test_node(topology, common_name=cert.common_name)
                    save.assert_called_once()
                    logger_exception.assert_called_once()
                    self.assertEqual(DeviceNode.objects.count(), 0)

    def test_filter_by_link(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')

        with self.subTest('empty result'):
            node = self._init_test_node(topology, common_name=cert.common_name)
            self.assertFalse(DeviceNode.filter_by_link(Link()).exists())

        with self.subTest('non empty result'):
            node2 = self._init_test_node(
                topology, addresses=['netjson_id2'], label='test2'
            )
            link = Link(
                source=node,
                target=node2,
                topology=topology,
                organization=topology.organization,
                cost=1,
            )
            link.full_clean()
            link.save()
            self.assertTrue(DeviceNode.filter_by_link(link).exists())

    def test_link_down_openvpn(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(topology, common_name=cert.common_name)
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        device.management_ip = '10.0.0.8'
        device.save()
        link = Link(
            source=node,
            target=node2,
            status='up',
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()
        with self.assertNumQueries(4):
            link.status = 'down'
            link.save()
        device.refresh_from_db()
        self.assertIsNone(device.management_ip)

    def test_link_up_openvpn(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(
            topology, addresses=['netjson_id', '10.0.0.2'], common_name=cert.common_name
        )
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        device.management_ip = None
        device.save()
        link = Link(
            source=node,
            target=node2,
            status='down',
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()
        with self.assertNumQueries(4):
            link.status = 'up'
            link.save()
        device.refresh_from_db()
        self.assertEqual(device.management_ip, '10.0.0.2')

    @mock.patch.object(models_logger, 'warning')
    def test_link_up_openvpn_failure(self, logger_warning):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(
            topology,
            addresses=['netjson_id', 'not_an_ip'],
            common_name=cert.common_name,
        )
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        device.management_ip = None
        device.save()
        link = Link(
            source=node,
            target=node2,
            status='down',
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()
        link.status = 'up'
        link.save()
        logger_warning.assert_called_once()
        logger_warning.assert_called_with(
            'ValueError raised while processing addresses: netjson_id, not_an_ip'
        )
        device.refresh_from_db()
        self.assertIsNone(device.management_ip)

    def test_node_label_override(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(topology, common_name=cert.common_name)
        self.assertEqual(node.get_name(), device.name)

    def test_topology_json(self):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        self._init_test_node(topology, common_name=cert.common_name)
        with self.subTest('assert number of queries'):
            with self.assertNumQueries(2):
                json = topology.json(dict=True)
        self.assertEqual(json['nodes'][0]['label'], device.name)

    def test_create_device_nodes_command(self):
        topology, _, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        properties = {'common_name': cert.common_name}
        n = self._create_node(topology=topology, properties=properties)
        DeviceNode.objects.all().delete()
        call_command('create_device_nodes')
        qs = DeviceNode.objects.filter(node=n)
        self.assertEqual(qs.count(), 1)


class TestMonitoringIntegration(Base, TransactionTestCase):
    @mock.patch('openwisp_monitoring.monitoring.apps.MonitoringConfig.create_database')
    @mock.patch('openwisp_monitoring.device.utils.manage_short_retention_policy')
    @mock.patch('openwisp_notifications.types.NOTIFICATION_TYPES', dict())
    def test_monitoring_integration(self, *args):
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node = self._init_test_node(
            topology, addresses=['netjson_id', '10.0.0.2'], common_name=cert.common_name
        )
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        device.management_ip = None
        device.save()
        link = Link(
            source=node,
            target=node2,
            status='up',
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()
        # needed for monitoring
        trigger_device_checks = import_string(trigger_device_checks_path)
        with mock.patch.object(trigger_device_checks, 'delay') as mocked_task:
            with self.modify_settings(
                INSTALLED_APPS={
                    'append': [
                        'openwisp_monitoring.monitoring',
                        'openwisp_monitoring.device',
                        'openwisp_monitoring.check',
                        'openwisp_controller.connection',
                        'openwisp_notifications',
                    ]
                }
            ):
                link.status = 'down'
                link.save()
            mocked_task.assert_called_once()
            mocked_task.assert_called_with(device.pk, recovery=False)


class TestAdmin(Base, TransactionTestCase):
    module = 'openwisp_network_topology'
    app_label = 'topology'
    topology_model = Topology
    link_model = Link
    node_model = Node
    user_model = get_user_model()
    fixtures = ['test_users.json']
    api_urls_path = 'api.urls'

    @property
    def prefix(self):
        return 'admin:{0}'.format(self.app_label)

    def setUp(self):
        org = self._create_org()
        t = self._create_topology(organization=org)
        topology, device, cert = self._create_test_env(parser='netdiff.OpenvpnParser')
        node1 = self._init_test_node(topology, common_name=cert.common_name)
        node2 = self._init_test_node(topology, addresses=['netjson_id2'], label='test2')
        link = Link(
            source=node1,
            target=node2,
            topology=topology,
            organization=topology.organization,
            cost=1,
        )
        link.full_clean()
        link.save()
        self._create_link(topology=t, source=node1, target=node2)
        self.client.force_login(self.user_model.objects.get(username='admin'))

    def test_node_change_list_queries(self):
        path = reverse('{0}_node_changelist'.format(self.prefix))
        with self.assertNumQueries(7):
            self.client.get(path)

    def test_link_change_list_queries(self):
        path = reverse('{0}_link_changelist'.format(self.prefix))
        with self.assertNumQueries(7):
            self.client.get(path)



import swapper
from django.http import HttpResponseForbidden
from django.shortcuts import render
from django.utils.translation import ugettext_lazy as _
from django.views import View

from ..settings import TOPOLOGY_API_AUTH_REQUIRED, VISUALIZER_CSS
from ..utils import get_object_or_404
from . import GraphVisualizerUrls

Topology = swapper.load_model('topology', 'Topology')


class TopologyListView(View):
    topology_model = Topology

    def get(self, request):
        topologies = self.topology_model.objects.filter(published=True)
        user = request.user
        if TOPOLOGY_API_AUTH_REQUIRED and not user.is_superuser:
            auth_perm = check_auth_perm(request)
            if auth_perm is not None:
                return auth_perm
            topologies = topologies.filter(organization__in=user.organizations_managed)
        return render(
            request,
            'netjsongraph/list.html',
            {'topologies': topologies, 'VISUALIZER_CSS': VISUALIZER_CSS},
        )


class TopologyDetailView(View, GraphVisualizerUrls):
    topology_model = Topology

    def get(self, request, pk):
        topology = get_object_or_404(self.topology_model, pk)
        user = request.user
        if TOPOLOGY_API_AUTH_REQUIRED and not user.is_superuser:
            auth_perm = check_auth_perm(request)
            if auth_perm is not None:
                return auth_perm
            if not user.is_manager(topology.organization):
                detail = _(
                    'User is not a manager of the organization to '
                    'which the requested resource belongs.'
                )
                return HttpResponseForbidden(detail)
        graph_url, history_url = self.get_graph_urls(request, pk)
        return render(
            request,
            'netjsongraph/detail.html',
            {
                'topology': topology,
                'graph_url': graph_url,
                'history_url': history_url,
                'VISUALIZER_CSS': VISUALIZER_CSS,
            },
        )


def check_auth_perm(request):
    user = request.user
    app_label = Topology._meta.app_label.lower()
    model = Topology._meta.object_name.lower()
    change_perm = f'{app_label}.change_{model}'
    view_perm = f'{app_label}.view_{model}'
    if not user.is_authenticated:
        return HttpResponseForbidden(_('Authentication credentials were not provided.'))
    if not (user.has_permission(change_perm) or user.has_permission(view_perm)):
        return HttpResponseForbidden(
            _('You do not have permission to perform this action.')
        )


topology_list = TopologyListView.as_view()
topology_detail = TopologyDetailView.as_view()
